// src/completion.builder.ts
import { json } from "@synstack/json";
import { pipe } from "@synstack/resolved";
import { t } from "@synstack/text";

// ../shared/src/ts.utils.ts
function never(neverCheck, value) {
  throw new Error(
    `Unexpected value, received: ${JSON.stringify(value ?? neverCheck, null, 2)}`
  );
}

// src/completion.runner.ts
import "@synstack/resolved";

// src/llm.types.ts
import "zod";

// src/completion.builder.ts
var CompletionBuilder2 = class _CompletionBuilder {
  _completion;
  constructor(completion2) {
    this._completion = completion2;
  }
  static get new() {
    return new _CompletionBuilder({});
  }
  static from(completion2) {
    return _CompletionBuilder.new.merge(completion2);
  }
  get $() {
    return this._completion;
  }
  $_(fn) {
    return pipe(this.$)._(fn);
  }
  async toSync() {
    return new _CompletionBuilder(await this.$);
  }
  /**
   * Override the values of the current builder with the provided values as object
   */
  merge(completion2) {
    return new _CompletionBuilder(
      pipe([this._completion, completion2])._(([old, added]) => ({
        ...old,
        ...added
      })).$
    );
  }
  // Todo: allow calling as a template directly
  /**
   * Set the system prompt
   */
  system(system) {
    return this.merge(
      pipe(system)._((system2) => ({
        system: system2
      })).$
    );
  }
  /**
   * Set the llm temperature
   * @argument temperature The temperature value between 0 and 1
   */
  temperature(temperature) {
    return this.merge(
      pipe(temperature)._((temperature2) => ({
        temperature: temperature2
      })).$
    );
  }
  /**
   * Set the maximum number of tokens to generate
   */
  maxTokens(maxTokens) {
    return this.merge(
      pipe(maxTokens)._((maxTokens2) => ({ maxTokens: maxTokens2 })).$
    );
  }
  /**
   * Set the stop sequences.
   *
   * If the LLM generates a stop sequence, the completion will stop.
   */
  stopSequences(sequences) {
    return this.merge(
      pipe(sequences)._((sequences2) => ({
        stopSequences: sequences2
      })).$
    );
  }
  /**
   * Set the top k tokens to consider for the completion
   */
  topK(topK) {
    return this.merge(
      pipe(topK)._((topK2) => ({
        topK: topK2
      })).$
    );
  }
  /**
   * Set the top p tokens to consider for the completion
   */
  topP(topP) {
    return this.merge(
      pipe(topP)._((topP2) => ({
        topP: topP2
      })).$
    );
  }
  /**
   * Reset the tools
   */
  clearTools() {
    return this.merge({
      toolsConfig: void 0
    });
  }
  /**
   * Set the tools available to the LLM
   *
   * @argument tools The tools to use
   * @argument requireToolUse Whether to require the LLM to use the tools
   */
  tools(tools, requireToolUse = false) {
    return this.merge(
      pipe(tools)._((tools2) => ({
        toolsConfig: {
          type: "multi",
          tools: tools2,
          requireToolUse
        }
      })).$
    );
  }
  /**
   * Set a single tool for the LLM to use. The LLM will always respond with this tool
   */
  tool(tool) {
    return this.merge(
      pipe(tool)._((tool2) => ({
        toolsConfig: {
          type: "single",
          tool: tool2
        }
      })).$
    );
  }
  /**
   * Set the stop reason for the completion
   */
  stopReason(stopReason) {
    return this.merge(
      pipe(stopReason)._((stopReason2) => ({
        stopReason: stopReason2
      })).$
    );
  }
  /**
   * Set the tokens usage
   */
  usage(usage) {
    return this.merge(
      pipe(usage)._((usage2) => ({
        usage: usage2
      })).$
    );
  }
  /**
   * Add usage to existing usage
   */
  addUsage(usage) {
    return this.merge(
      pipe([this.$, usage])._(([$, usage2]) => ({
        usage: {
          inputTokens: $.usage?.inputTokens ?? 0 + usage2.inputTokens,
          outputTokens: $.usage?.outputTokens ?? 0 + usage2.outputTokens
        }
      })).$
    );
  }
  /**
   * Set messages
   */
  messages(messages) {
    return this.merge(
      pipe(messages)._((messages2) => ({
        messages: messages2
      })).$
    );
  }
  /**
   * Add messages
   */
  addMessages(messages) {
    return this.merge(
      pipe([this.$, messages])._(([$, messages2]) => ({
        messages: [...$.messages ?? [], ...messages2]
      })).$
    );
  }
  /**
   * Add assistant message
   */
  addAssistantMessage(message) {
    return this.merge(
      pipe([this.$, message])._(([$, message2]) => {
        const $messages = $.messages ?? [];
        const lastMessage = $.messages.at(-1);
        if (!lastMessage)
          return {
            messages: [message2]
          };
        if (lastMessage.role !== "assistant")
          return {
            messages: [...$messages, message2]
          };
        return {
          messages: [
            ...$messages.slice(0, -1),
            {
              ...lastMessage,
              content: [...lastMessage.content, ...message2.content]
            }
          ]
        };
      }).$
    );
  }
  /**
   * Execute completion on Llm runner
   */
  run(runner) {
    return runner.runChatCompletion(this);
  }
  lastMessage() {
    return pipe(this.$)._((completion2) => {
      const lastMessage = completion2.messages?.at(-1);
      if (!lastMessage) throw new EmptyMessagesException(completion2);
      return lastMessage;
    }).$;
  }
  /**
   * Get last message as text
   */
  lastMessageText() {
    return pipe(this.lastMessage())._(
      (message) => message.content.filter((c) => c.type === "text").map((c) => c.text).join("")
    ).$;
  }
  /**
   * Get last message tool calls
   */
  lastMessageToolCalls() {
    return pipe(this.lastMessage())._((message) => {
      if (message.role !== "assistant") return [];
      return message.content.filter((c) => c.type === "tool_call");
    }).$;
  }
};
var EmptyMessagesException = class extends Error {
  constructor(completion2) {
    super(t`
      No messages in current completion state:
        ${json.serialize(completion2, { pretty: true })} 
    `);
  }
};
var completion = CompletionBuilder2.new;

// src/message.builder.ts
import { pipe as pipe2 } from "@synstack/resolved";

// ../fs/src/file.lib.ts
import { glob as glob2 } from "@synstack/glob";
import { json as json2 } from "@synstack/json";
import { path as path2 } from "@synstack/path";
import { Pipeable as Pipeable2 } from "@synstack/pipe";
import { str } from "@synstack/str";
import { xml } from "@synstack/xml";
import { yaml } from "@synstack/yaml";
import * as fsSync2 from "fs";
import * as fs2 from "fs/promises";
import "zod";

// ../fs/src/dir.lib.ts
import { git } from "@synstack/git";
import { glob } from "@synstack/glob";
import { path } from "@synstack/path";
import { Pipeable } from "@synstack/pipe";
import * as fsSync from "fs";
import * as fs from "fs/promises";

// ../fs/src/files-array.lib.ts
import { enhance } from "@synstack/enhance";
import "@synstack/path";
var filesArrayMethods = {
  filter(fn) {
    return files(this.$.filter(fn));
  },
  filterGlobs(...patterns) {
    return this.filter((file2) => file2.matchesGlobs(...patterns));
  },
  filterMimeTypes(...mimeTypes) {
    const types = new Set(mimeTypes);
    return this.filter((file2) => {
      const mimeType = file2.mimeType();
      if (mimeType === null) return false;
      return types.has(mimeType);
    });
  },
  filterDir(dirOrPath) {
    return this.filter((file2) => file2.isInDir(dirOrPath));
  },
  toPaths() {
    return this.map((file2) => file2.path);
  },
  relativePathsTo(dirOrFileOrPath) {
    return this.map((file2) => file2.relativePathTo(dirOrFileOrPath));
  }
};
var files = (files2) => enhance(
  "files_array",
  files2.map((f) => {
    if (f instanceof FsFile) return f;
    return file(f);
  }),
  filesArrayMethods
);
var filesFromDir = (dir2) => (files2) => enhance(
  "files_array",
  files2.map((f) => {
    if (f instanceof FsFile) return f;
    return dir2.file(f);
  }),
  filesArrayMethods
);

// ../fs/src/dir.lib.ts
var FsDir2 = class _FsDir extends Pipeable {
  _path;
  constructor(path3) {
    super();
    this._path = path3;
  }
  /**
   * @returns The absolute path of the directory
   */
  toString() {
    return this._path;
  }
  /**
   * @returns The absolute path of the directory
   */
  valueOf() {
    return this._path;
  }
  instanceOf() {
    return this;
  }
  /**
   * @returns The absolute path of the directory
   */
  get path() {
    return this._path;
  }
  /**
   * Create a new directory instance with the provided path
   */
  static cwd(...paths) {
    return new _FsDir(path.resolve(...paths));
  }
  /**
   * Generates a new FsDir instance by joining the given relative path
   * to the current directory's path
   *
   * @param relativePath - The relative path to append to the current directory
   * @returns A new FsDir instance representing the combined path
   */
  to(relativePath) {
    const newPath = path.join(this._path, relativePath);
    return new _FsDir(newPath);
  }
  /**
   * Generates a new FsFile instance by joining the given relative path
   * to the current directory's path
   *
   * @param relativePath - The relative path to append to the current directory.
   * @returns A new FsFile instance representing the combined path.
   */
  file(relativePath) {
    if (path.isAbsolute(relativePath))
      throw new Error(`
Trying to access a dir file from an absolute paths:
  - Folder path: ${this._path}
  - File path: ${relativePath}
`);
    return FsFile.from(this._path, relativePath);
  }
  /**
   * @returns The name of the directory
   */
  name() {
    return path.filename(this._path);
  }
  /**
   * Checks if the directory exists
   */
  async exists() {
    return fs.access(this._path, fsSync.constants.F_OK).then(() => true).catch(() => false);
  }
  /**
   * Checks if the directory exists
   * @synchronous
   */
  existsSync() {
    try {
      fsSync.accessSync(this._path, fsSync.constants.F_OK);
      return true;
    } catch (error) {
      if (error.code === "ENOENT") return false;
      throw error;
    }
  }
  /**
   * Creates the directory if it doesn't exist
   */
  async make() {
    await fs.mkdir(this._path, { recursive: true });
  }
  /**
   * Creates the directory if it doesn't exist
   * @synchronous
   */
  makeSync() {
    fsSync.mkdirSync(this._path, { recursive: true });
  }
  /**
   * Removes the directory and all its contents recursively
   */
  async rm() {
    await fs.rm(this._path, { recursive: true }).catch((e) => {
      if (e.code === "ENOENT") return;
      throw e;
    });
  }
  /**
   * Removes the directory and all its contents recursively
   * @synchronous
   */
  rmSync() {
    try {
      fsSync.rmSync(this._path, { recursive: true });
    } catch (error) {
      if (error.code === "ENOENT") return;
      throw error;
    }
  }
  /**
   * Runs a glob search on the directory and returns the matching files
   * @returns an FsFileArray instance with the matching files
   */
  async glob(...patterns) {
    return glob.cwd(this._path).find(...patterns).then(filesFromDir(this));
  }
  /**
   * Runs a glob search on the directory and returns the matching files
   * @returns an FsFileArray instance with the matching files
   * @synchronous
   */
  globSync(...patterns) {
    return filesFromDir(this)(glob.cwd(this._path).findSync(...patterns));
  }
  /**
   * Runs a git ls-files search on the directory and returns the matching files
   * @param subPath a sub path to be appended to the git ls-files search
   * @returns an FsFileArray instance with the matching files
   */
  async gitLs(subPath) {
    return git.ls(this._path, subPath).then(files);
  }
};
var dir = FsDir2.cwd;

// ../fs/src/file.lib.ts
var FsFile = class _FsFile extends Pipeable2 {
  _path;
  _encoding;
  _schema;
  /**
   * Create a new FsFile instance from a path or a list of paths to be resolved
   * Path will be an absolute path
   */
  static from(...paths) {
    return new _FsFile(path2.join(...paths), "utf-8");
  }
  constructor(path3, encoding, schema) {
    super();
    this._path = path3;
    this._encoding = encoding ?? "utf-8";
    this._schema = schema ?? void 0;
  }
  /**
   * Provide a validation schema for the file content. To be used with:
   * - `.read.json`
   * - `.write.json`
   * - `.read.yaml`
   * - `.write.yaml`
   */
  schema(schema) {
    return new _FsFile(this._path, this._encoding, schema);
  }
  /**
   * @returns The path of the file
   */
  valueOf() {
    return this._path;
  }
  instanceOf() {
    return this;
  }
  // #region sub actions
  /**
   * Access the read actions of the file
   */
  get read() {
    return new FsFileRead(
      this._path,
      this._encoding,
      this._schema
    );
  }
  /**
   * Access the write actions of the file
   */
  get write() {
    return new FsFileWrite(
      this._path,
      this._encoding,
      "overwrite",
      this._schema
    );
  }
  // #endregion
  // #region sync
  /**
   * @returns The absolute path of the file
   */
  get path() {
    return this._path;
  }
  /**
   * @returns The absolute path of the directory containing the file
   */
  dirPath() {
    return path2.dirname(this._path);
  }
  /**
   * @returns An FsDir instance representing the directory containing the file
   */
  dir() {
    return FsDir2.cwd(this.dirPath());
  }
  /**
   * @returns The name of the file
   */
  fileName() {
    return path2.filename(this._path);
  }
  /**
   * @returns The extension of the file
   */
  fileExtension() {
    return path2.fileExtension(this._path);
  }
  /**
   * @returns The name of the file without the extension
   */
  fileNameWithoutExtension() {
    return path2.filenameWithoutExtension(this._path);
  }
  /**
   * @returns The mime type of the file or null if it cannot be determined
   */
  mimeType() {
    return path2.mimeType(this._path);
  }
  /**
   * Create a new FsFile instance with the provided relative path
   * @argument relativePath The relative path of the new file
   */
  toFile(relativePath) {
    const newPath = path2.resolve(this.dirPath(), relativePath);
    return new _FsFile(newPath);
  }
  /**
   * Create a new FsDir instance with the provided relative path
   * @argument relativePath The relative path of the new directory
   */
  toDir(relativePath) {
    const newPath = path2.join(this.dirPath(), relativePath);
    return FsDir2.cwd(newPath);
  }
  /**
   * Get the relative path from another file to this file
   *
   * ```ts
   * import { file } from "@synstack/fs";
   *
   * const file1 = file("/path/to/file1.txt");
   * const file2 = file("/path/to-other/file2.txt");
   *
   * console.log(file1.relativePathFrom(file2)); // ../to/file1.txt
   * ```
   */
  relativePathFrom(dirOrFileOrPath) {
    if (dirOrFileOrPath instanceof _FsFile)
      return this.relativePathFrom(dirOrFileOrPath.dir());
    return path2.relative(dirOrFileOrPath.path, this.path);
  }
  /**
   * Get the relative path to go from this file to another
   *
   * ```ts
   * import { file } from "@synstack/fs";
   *
   * const file1 = file("/path/to/file1.txt");
   * const file2 = file("/path/to-other/file2.txt");
   *
   * console.log(file1.relativePathTo(file2)); // ../to-other/file2.txt
   * ```
   */
  relativePathTo(dirOrFileOrPath) {
    return path2.relative(this.dirPath(), dirOrFileOrPath.path);
  }
  /**
   * Check if the file is in the provided directory
   */
  isInDir(dirOrPath) {
    return path2.isInPath(dirOrPath.valueOf(), this._path.valueOf());
  }
  /**
   * Delete the file from the file system
   */
  async rm() {
    await fs2.rm(this._path, { recursive: true }).catch((e) => {
      if (e.code === "ENOENT") return;
      throw e;
    });
  }
  /**
   * Delete the file from the file system
   * @synchronous
   */
  rmSync() {
    try {
      fsSync2.rmSync(this._path, { recursive: false });
    } catch (error) {
      if (error.code === "ENOENT") return;
      throw error;
    }
  }
  /**
   * Check if the file exists
   */
  async exists() {
    return fs2.access(this._path, fs2.constants.F_OK).then(() => true).catch(() => false);
  }
  /**
   * Check if the file exists
   * @synchronous
   */
  existsSync() {
    try {
      fsSync2.accessSync(this._path);
      return true;
    } catch (error) {
      if (error.code === "ENOENT") return false;
      throw error;
    }
  }
  /**
   * Get the creation date of the file as a Date object
   */
  async creationDate() {
    const fileStats = await fs2.stat(this._path);
    return fileStats.birthtime;
  }
  /**
   * Get the creation date of the file as a Date object
   * @synchronous
   */
  creationDateSync() {
    const stats = fsSync2.statSync(this._path);
    return stats.birthtime;
  }
  /**
   * Check if the file matches any of the glob patterns
   */
  matchesGlobs(...globs) {
    return glob2.matches(this._path, ...globs);
  }
  /**
   * Capture parts of the file path using a glob pattern
   *
   * ```ts
   * import { file } from "@synstack/fs";
   *
   * const myFile = file("/my-domain/my-sub-domain/features/feature-name.controller.ts");
   * const res = myFile.globCapture("/(*)/(*)/features/(*).controller.ts");
   * if (!res) throw new Error("File doesn't match glob pattern");
   * console.log(res[1]); // my-domain
   * console.log(res[2]); // my-sub-domain
   * console.log(res[3]); // feature-name.controller.ts
   * ```
   */
  globCapture(globPattern) {
    return glob2.capture(globPattern, this._path);
  }
};
var FsFileRead = class {
  _path;
  _encoding;
  _schema;
  constructor(path3, encoding, schema) {
    this._path = path3;
    this._encoding = encoding;
    this._schema = schema;
  }
  get path() {
    return this._path;
  }
  // #region sync
  /**
   * @returns The content of the file as a string
   */
  async text() {
    return fs2.readFile(this._path, this._encoding);
  }
  /**
   * @returns The content of the file as a string
   * @synchronous
   */
  textSync() {
    return fsSync2.readFileSync(this._path, this._encoding);
  }
  /**
   * Used for further manipulation of the content
   * @returns The content of the file as a synstack/str instance
   */
  async str() {
    return this.text().then(str);
  }
  /**
   * Used for further manipulation of the content
   * @returns The content of the file as a synstack/str instance
   * @synchronous
   */
  strSync() {
    return str(this.textSync());
  }
  /**
   * @returns The content of the file as a JSON object
   * @info If a schema was provided to the file instanse, it will be used to validate the JSON
   */
  async json() {
    return this.text().then(
      (t3) => json2.deserialize(t3, { schema: this._schema })
    );
  }
  /**
   * @returns The content of the file as a JSON object synchronously
   * @info If a schema was provided to the file instanse, it will be used to validate the JSON
   */
  jsonSync() {
    return json2.deserialize(this.textSync(), { schema: this._schema });
  }
  /**
   * @returns The YAML content of the file as a deserialized object
   * @info If a schema was provided to the file instanse, it will be used to validate the data
   */
  async yaml() {
    return this.text().then(
      (t3) => yaml.deserialize(t3, { schema: this._schema })
    );
  }
  /**
   * @returns The YAML content of the file as a deserialized object synchronously
   * @info If a schema was provided to the file instanse, it will be used to validate the data
   */
  yamlSync() {
    return yaml.deserialize(this.textSync(), { schema: this._schema });
  }
  /**
   * Uses synstack/xml to parse the file content
   *
   * - Synstack XML is a non spec-compliant XML parser tailored for LLMs
   * - Check it's [documentation](https://github.com/pAIrprogio/synscript/tree/main/packages/xml) for more information
   */
  async xml() {
    return this.text().then(xml.parse);
  }
  /**
   * Uses synstack/xml to parse the file content synchronously
   *
   * - Synstack XML is a non spec-compliant XML parser tailored for LLMs
   * - Check it's [documentation](https://github.com/pAIrprogio/synscript/tree/main/packages/xml) for more information
   */
  xmlSync() {
    return xml.parse(this.textSync());
  }
  /**
   * @returns The content of the file as a base64 string
   */
  async base64() {
    return fs2.readFile(this._path, "base64");
  }
  /**
   * @returns The content of the file as a base64 string synchronously
   */
  base64Sync() {
    return fsSync2.readFileSync(this._path, "base64");
  }
  /**
   * @returns The content of the file as a synstack compatible Base64 data object
   */
  async base64Data(defaultMimeType = "application/octet-stream") {
    return {
      type: "base64",
      data: await this.base64(),
      mimeType: defaultMimeType
    };
  }
  /**
   * @returns The content of the file as a synstack compatible Base64 data object synchronously
   */
  base64DataSync(defaultMimeType = "application/octet-stream") {
    return {
      type: "base64",
      data: this.base64Sync(),
      mimeType: defaultMimeType
    };
  }
};
var FsFileWrite = class _FsFileWrite {
  _path;
  _encoding;
  _schema;
  _mode;
  constructor(path3, encoding, mode = "overwrite", schema) {
    this._path = path3;
    this._encoding = encoding;
    this._schema = schema;
    this._mode = mode;
  }
  /**
   * Set the write mode of the file
   * @argument preserve: If the file already exists, it will be left untouched
   * @argument overwrite: If the file already exists, it will be overwritten
   */
  mode(writeMode) {
    return new _FsFileWrite(this._path, this._encoding, writeMode, this._schema);
  }
  /**
   * Write text to the file
   */
  async text(content) {
    if (this._mode === "preserve" && await FsFile.from(this._path).exists())
      return;
    const dirname = path2.dirname(this._path);
    await fs2.mkdir(dirname, { recursive: true });
    await fs2.writeFile(this._path, content.toString(), this._encoding);
  }
  /**
   * Write text to the file synchronously
   */
  textSync(content) {
    if (this._mode === "preserve" && FsFile.from(this._path).existsSync())
      return;
    const dirname = path2.dirname(this._path);
    fsSync2.mkdirSync(dirname, { recursive: true });
    fsSync2.writeFileSync(this._path, content.toString(), this._encoding);
  }
  /**
   * Write JSON to the file
   * @argument data The data to write, can be any JSON-serializable value
   */
  async json(data) {
    return this.text(json2.serialize(data, { schema: this._schema }));
  }
  /**
   * Write pretty JSON to the file
   * @argument data The data to write, can be any JSON-serializable value
   */
  async prettyJson(data) {
    return this.text(
      json2.serialize(data, { schema: this._schema, pretty: true }) + "\n"
    );
  }
  /**
   * Write JSON to the file
   * @argument data: The data to write, can be any JSON-serializable value
   * @info If a schema was provided to the file instanse, it will be used to validate the data before writing
   * @synchronous
   */
  jsonSync(data) {
    return this.textSync(json2.serialize(data, { schema: this._schema }));
  }
  /**
   * Write pretty JSON to the file
   * @argument data: The data to write, can be any JSON-serializable value
   * @info If a schema was provided to the file instanse, it will be used to validate the data before writing
   * @synchronous
   */
  prettyJsonSync(data) {
    return this.textSync(
      json2.serialize(data, { schema: this._schema, pretty: true }) + "\n"
    );
  }
  /**
   * Write YAML to the file
   * @argument data The data to write, can be any JSON-serializable value
   * @info If a schema was provided to the file instanse, it will be used to validate the data
   */
  async yaml(data) {
    return this.text(yaml.serialize(data, { schema: this._schema }));
  }
  /**
   * Write YAML to the file
   * @argument data The data to write, can be any JSON-serializable value
   * @info If a schema was provided to the file instanse, it will be used to validate the data
   * @synchronous
   */
  yamlSync(data) {
    return this.textSync(yaml.serialize(data, { schema: this._schema }));
  }
  /**
   * Write Base64 to the file
   * @argument data The data to write, needs to be a base64 string
   */
  async base64(data) {
    if (this._mode === "preserve" && await FsFile.from(this._path).exists())
      return;
    const dirname = path2.dirname(this._path);
    await fs2.mkdir(dirname, { recursive: true });
    await fs2.writeFile(this._path, data.toString(), "base64");
  }
  /**
   * Write Base64 to the file
   * @argument data The data to write, needs to be a base64 string
   * @synchronous
   */
  base64Sync(data) {
    if (this._mode === "preserve" && FsFile.from(this._path).existsSync())
      return;
    const dirname = path2.dirname(this._path);
    fsSync2.mkdirSync(dirname, { recursive: true });
    return fsSync2.writeFileSync(this._path, data.toString(), "base64");
  }
};
var file = FsFile.from;

// src/message.builder.ts
import { t as t2, tParse } from "@synstack/text";
var userMsg = (template, ...values) => pipe2(t2(template, ...values))._(tParse)._(
  (parts) => parts.map((v) => {
    if (typeof v === "string")
      return { type: "text", text: v };
    if (v.type === "base64")
      return {
        type: "image",
        image: v
      };
    if (v.type === "tool_response") return v;
    never(v);
  })
)._(
  (content) => ({
    role: "user",
    content
  })
).$;
var assistantMsg = (template, ...values) => {
  return pipe2(t2(template, ...values))._(tParse)._(
    (parts) => parts.map((v) => {
      if (typeof v === "string")
        return { type: "text", text: v };
      if (v.type === "tool_call") return v;
      never(v.type);
    })
  )._(
    (content) => ({
      role: "assistant",
      content
    })
  ).$;
};
export {
  CompletionBuilder2 as CompletionBuilder,
  assistantMsg,
  completion,
  userMsg
};
//# sourceMappingURL=llm.index.js.map